@@ -2,12 +2,20 @@
 '''
 Created on Feb 2, 2014
 
+Given a course name, compute the student time-on-task (engagement).
+Two variants: can use 'All' for course name to have all engagements
+computed. Secondly: can add any number of years. When provided, 
+only courses whose first event was in any one of those years will
+be processed. 
+
 Can be called from commandline, or imported into another module, which
 does the little setup in __main__ below, and then invokes run() and
-writeToDisk() on an instance of this class. A convenience script 
-scripts/computeEngagement.sh is available, but not necessary.
+writeToDisk() on an instance of this class. See __main__ for options.
+See open_edx_export_class/src/exportClass.py for example of using
+this module as a library.
 
-Assumes availability of the following DB table:
+Assumes availability of the following DB table, which is generated 
+via script prepEngagementAnalysis.sh 
 mysql> DESCRIBE Activities;
 +---------------------+--------------+------+-----+---------------------+-------+
 | Field               | Type         | Null | Key | Default             | Extra |
@@ -25,35 +33,8 @@ Example rows:
 | Education/EDUC115N/How_to_Learn_Math | 00014bffc716bf9d8d656d2f668f737cd43acde8 | play_video | 2013-07-20 22:59:36      | 1 |
 
 
-Also assumes availability of the following DB table;
-See scripts/prepEngagementAnalysis.sh for how to build
-that table:
-
-mysql> DESCRIBE CourseRuntimes;
-+---------------------+--------------+------+-----+---------+-------+
-| Field               | Type         | Null | Key | Default | Extra |
-+---------------------+--------------+------+-----+---------+-------+
-| course_display_name | varchar(255) | YES  |     | NULL    |       |
-| course_start_date   | datetime     | YES  |     | NULL    |       |
-| course_end_date     | datetime     | YES  |     | NULL    |       |
-+---------------------+--------------+------+-----+---------+-------+
-
-Example rows:
-mysql> SELECT * FROM CourseRuntimes;
-+-------------------------------------------------------+---------------------+---------------------+
-| course_display_name                                   | course_start_date   | course_end_date     |
-+-------------------------------------------------------+---------------------+---------------------+
-| Medicine/SciWrite/Fall2013                            | 2013-11-10 06:45:16 | 2013-11-17 06:56:35 |
-| Engineering/EE-222/Applied_Quantum_Mechanics_I        | 2013-11-10 06:54:21 | 2013-11-17 06:54:21 |
-| Engineering/Solar/Fall2013                            | 2013-11-10 06:44:13 | 2013-11-17 07:00:26 |
-| Engineering/CS144/Introduction_to_Computer_Networking | 2013-11-10 06:52:22 | 2013-11-17 06:58:49 |
-+-------------------------------------------------------+---------------------+---------------------+
-4 rows in set (0.00 sec)
-
-Grouped by course, then student, and ordered by time.
-The course end time is only used to filter
-out courses that seem to have lasted less than 7 days.
-Those courses tend to be test courses.
+To use, instantiate EngagementComputer, call the run() method,
+and then the writeResultsToDisk() method.   
 
 The output are three files: /tmp/engagement.log,  /tmp/engagementAllCourses_summary.csv,
 and /tmp/engagementAllCourses_allData.csv. The summary file:
@@ -85,6 +66,8 @@ The engagementAllCourses_allData.csv contains every session of every student.
 
 @author: paepcke
 '''
+import argparse
+import copy
 import datetime
 import getpass
 import math
@@ -96,10 +79,10 @@ import sys
 import tempfile
 import time
 
-#from mysqldb import MySQLDB
 from pymysql_utils.pymysql_utils import MySQLDB
 
 
+#from mysqldb import MySQLDB
 # Add json_to_relation source dir to $PATH
 # for duration of this execution:
 source_dir = [os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../json_to_relation/json_to_relation")]
@@ -110,21 +93,27 @@ sys.path = source_dir
 
 class EngagementComputer(object):
     
-    # Time in minutes within which a student activity event
-    # must follow her previous event to be counted
-    # as having occurred within one session:
-    SESSION_TIMEOUT = 3 * 60
-    
     # Time duration counted for any video event (minutes):
     VIDEO_EVENT_DURATION = 15
     
     # Time duration counted for any non-video event (minutes):
     NON_VIDEO_EVENT_DURATION = 5
     
+    # Database that contains EventXtract table:
+    EVENT_XTRACT_TABLE_DB = 'Edx'
+    
     # Recognizing fake course names:
     FAKE_COURSE_PATTERN = re.compile(r'([Tt]est|[Ss]and[Bb]ox|[Dd]avid|[Dd]emo|Humaanities|SampleUniversity|[Jj]ane|ZZZ|Education/EDUC115N[^\s]*\s)')
     
-    def __init__(self, coursesStartYearsArr, dbHost, dbName, tableName, openMySQLDB=None, mySQLUser=None, mySQLPwd=None, courseToProfile=None):
+    def __init__(self, 
+                coursesStartYearsArr, 
+                dbHost, 
+                dbName, 
+                tableName, 
+                mySQLUser=None, 
+                mySQLPwd=None, 
+                courseToProfile=None, 
+                sessionInactivityThreshold=30):
         '''
         Sets up one session-accounting run through a properly filled table (as
         per file level comment above.
@@ -136,7 +125,7 @@ class EngagementComputer(object):
         @type dbHost: string
         @param dbName: name of db within server in which the activities table resides. Use 
             this parameter to place test tables into, say the 'test' database. Point this dbName
-            parameter to 'test' and all ops will look for the activities table and CourseRuntimes
+            parameter to 'test' and all ops will look for the Activities table and CourseRuntimes
             table in that db.
         @type dbName: string
         @param tableName: name of table that holds the activities as per file level header. 
@@ -148,15 +137,23 @@ class EngagementComputer(object):
         @param courseToProfile: name of course to analyze for sessions. If None all courses 
              that started in one of the years listed in coursesStartYearArr will be examined. 
         @type courseToProfile: [string]
+        @param sessionInactivityThreshold: time in minutes of student inactivity beyond which 
+               it is concluded that the student is no longer working on the computer in the
+               current session.
+        @type sessionInactivityThreshold: int
         '''
         self.dbHost = dbHost
         self.dbName = dbName
         self.tableName = tableName
         self.mySQLUser = mySQLUser
         self.mySQLPwd  = mySQLPwd
-        self.db        = openMySQLDB
-        self.courseToProfile = courseToProfile
+        if courseToProfile == "None":
+            self.courseToProfile = None
+        else:
+            self.courseToProfile = courseToProfile
+        
         self.coursesStartYearsArr = coursesStartYearsArr
+        self.sessionInactivityThreshold = sessionInactivityThreshold
         # To keep track of which course runtimes
         # we already output an error msg for,
         # b/c we didn't find it:
@@ -175,14 +172,7 @@ class EngagementComputer(object):
                 self.mySQLPwd = ''
         # Place to hold all stats for one class
         self.classStats = {}
-        # If no open db connector was passed in, open one:
-        if self.db is None:
-            self.db = MySQLDB(host=self.dbHost, user=self.mySQLUser, passwd=self.mySQLPwd, db=dbName)
-            self.openOwnDB = True
-        else:
-            self.openOwnDB = False
-        # Ensure that the CourseRuntimes table exists:
-        self.getCourseRuntime('fakename', testOnly=True)        
+        self.db = MySQLDB(host=self.dbHost, user=self.mySQLUser, passwd=self.mySQLPwd, db=dbName)
         
     def run(self):
         '''
@@ -216,11 +206,11 @@ class EngagementComputer(object):
             queryEndTimeReported = False
             if self.courseToProfile is None:
                 # Profile all courses:
-                queryIterator = self.db.query('use %s; SELECT course_display_name, anon_screen_name, time, isVideo FROM %s ORDER BY course_display_name, anon_screen_name, time;' % 
-                                              self.dbName, self.tableName)
+                queryIterator = self.db.query('SELECT course_display_name, anon_screen_name, time, isVideo FROM %s ORDER BY course_display_name, anon_screen_name, time;' %
+                                              self.tableName)
             else:
-                queryIterator = self.db.query('use %s; SELECT course_display_name, anon_screen_name, time, isVideo FROM %s WHERE course_display_name LIKE "%s" ORDER BY anon_screen_name, time;' % 
-                                              (self.dbName, self.tableName, self.courseToProfile))
+                queryIterator = self.db.query('SELECT course_display_name, anon_screen_name, time, isVideo FROM %s WHERE course_display_name = "%s" ORDER BY anon_screen_name, time;' %\
+                                              (self.tableName, self.courseToProfile))
             for activityRecord in queryIterator:
                 if not queryEndTimeReported:
                     self.log('Query done in %s' % str(datetime.timedelta(seconds=(time.time() - queryStartTime))))
@@ -241,20 +231,25 @@ class EngagementComputer(object):
                        len(currEvent['course_display_name']) > 0:
                         self.currStudent = currEvent['anon_screen_name']
                         self.currCourse  = currEvent['course_display_name']
-                        # Get start and end dates of this class:
-                        try:
-                            (self.courseStartDate, self.courseEndDate) = self.getCourseRuntime(self.currCourse)
-                        except Exception as e:
-                            self.logErr("While calling getCourseRuntime() from run(): '%s'" % `e`)
-                            continue
-                        # If getCourseRuntime() failed, that method will have logged
-                        # the error, so we just continue:
-                        if self.courseStartDate is None or self.courseEndDate is None:
-                            continue
-                        # Only deal with classes that started in the 
-                        # desired year:
-                        if not self.courseStartDate.year in self.coursesStartYearsArr:
-                            continue
+                        # If we are only to consider courses that started during
+                        # a particular year, then find this course's start/end times,
+                        # and ignore the course if it's not in one of the acceptable
+                        # years:
+                        if self.coursesStartYearsArr is not None:
+                            # Get start and end dates of this class:
+                            try:
+                                (self.courseStartDate, self.courseEndDate) = self.getCourseRuntime(self.currCourse)
+                            except Exception as e:
+                                self.logErr("While calling getCourseRuntime() from run(): '%s'" % `e`)
+                                continue
+                            # If getCourseRuntime() failed, that method will have logged
+                            # the error, so we just continue:
+                            if self.courseStartDate is None or self.courseEndDate is None:
+                                continue
+                            # Only deal with classes that started in the 
+                            # desired year:
+                            if not self.courseStartDate.year in self.coursesStartYearsArr:
+                                continue
                         self.sessionStartTime = currEvent['eventDateTime']
                         prevEvent = currEvent
                         self.log("Starting on course %s..." % currEvent['course_display_name'])
@@ -307,11 +302,7 @@ class EngagementComputer(object):
         finally:
             if self.db is not None:
                 try:
-                    # If creator of this instance passed an open
-                    # db into this instance, don't close that db.
-                    # But if we opened one ourselves, close it:
-                    if self.openOwnDB:
-                        self.db.close()
+                    self.db.close()
                 except Exception as e:
                     self.logErr('Could not close activities db: ' % `e`);
 
@@ -319,7 +310,7 @@ class EngagementComputer(object):
         '''
         Called when a new event by a student is being processed. Adds the
         event time to self.timeSpentThisSession. If time since previous
-        event > SESSION_TIMEOUT, the current session is finalized, and
+        event > self.sessionInactivityThreshold, the current session is finalized, and
         a new session is started.
         @param dateTimePrevEvent:
         @type dateTimePrevEvent:
@@ -334,7 +325,7 @@ class EngagementComputer(object):
             self.sessionStartTime = dateTimeCurrEvent
         timeDelta = dateTimeCurrEvent - dateTimePrevEvent
         minutes   = round(timeDelta.total_seconds()/60.0)
-        if minutes > EngagementComputer.SESSION_TIMEOUT:
+        if minutes > self.sessionInactivityThreshold:
             self.wrapUpSession(self.currStudent, isVideo, timeSpentSoFar, dateTimeCurrEvent)
         else:
             newTimeSpent = timeSpentSoFar + minutes
@@ -356,7 +347,7 @@ class EngagementComputer(object):
     
     def wrapUpSession(self, currentStudent, wasVideo, timeSpentSoFar, dateTimeNewSessionStart):
         '''
-        A student event is more than SESSION_TIMEOUT after the previous
+        A student event is more than self.sessionInactivityThreshold after the previous
         event by the same student in the same class.
         @param currentStudent: student who is currently under analysis
         @type currentStudent: string
@@ -385,9 +376,16 @@ class EngagementComputer(object):
         Called when all students of one class have been
         processed. This method receives a dict that maps
         each student of the class being closed out to 
-        an array of that student's session time.
-        
-        
+        an array of that student's session times and lenghts. 
+        For each student session, the dict value is an
+        array of two-tuples: (sessionStartTime,sessionLength).
+        The studentSessionsDict thus looks like this:
+            {student1 : [(firstTimeS1, 10), (secondTimeS1, 4), ...]
+             student2 : [(firstTimeS2, 10), (secondTimeS2, 4), ...]
+             
+        Important: The times within each array are sorted, so
+        sessionStartTime_n+1 > sessionStartTime_n.
+        We take advantage of this fact to optimize.
         @param studentSessionsDict:
         @type studentSessionsDict:
         '''
@@ -419,22 +417,63 @@ class EngagementComputer(object):
             totalEffortAllStudents = 0
             totalStudentSessions    = 0
             
-            for weekNum in range(numWeeks):
+            studentSessionsLeftToDo = {}
+            
+            for weekNum in range(numWeeks+1):
                 weekStart = startDate + weekNum * datetime.timedelta(weeks=1)
                 weekEnd   = weekStart + datetime.timedelta(weeks=1)
                 for student in self.studentSessionsDict.keys():
                     thisWeekThisStudentSessionList = []
-                    dateAndSessionLenArr = self.studentSessionsDict[student]
-                    for (eventDateTime, engageDurationMins) in dateAndSessionLenArr:
+                    
+                    try:
+                        # If we looked at this student before (i.e. during
+                        # an earlier week, retrieve the array of sessionTime/Len
+                        # two-tuples that are left to process: 
+                        dateAndSessionLenArr = studentSessionsLeftToDo[student]
+                    except KeyError:
+                        # First time we see this student. 
+                        # Get a *copy* of this student's array of time-sorted sessionTime/sessionLen tuples.
+                        # The copying takes time, but working on a copy
+                        # we allows us to delete elements we are done with
+                        # for the purpose of this loop, rather than having to
+                        # go through all the elements left to right each time
+                        # through the loop below:
+                        dateAndSessionLenArr = copy.copy(self.studentSessionsDict[student])
+                        studentSessionsLeftToDo[student] = dateAndSessionLenArr
+
+                    if len(dateAndSessionLenArr) == 0:
+                        continue
+                    while True: 
+                        try:
+                            # Always just look at first sessionTime/Len tuple,
+                            # because we remove the front of the array each time around:
+                            (eventDateTime, engageDurationMins) = dateAndSessionLenArr[0]
+                        except IndexError:
+                            # Done with this student
+                            break
+                        
                         if not isinstance(eventDateTime, datetime.datetime):
                             self.logErr("Expected datetime, but got %s ('%s') from dateAndSessionLenArr." % 
                                              (type(eventDateTime), str(eventDateTime)))
+                            # Don't want to see this array entry again:
+                            dateAndSessionLenArr.pop(0)
                             continue
-                        if eventDateTime < weekStart or\
-                           eventDateTime > weekEnd or\
-                           engageDurationMins == 0:
+                        if eventDateTime < weekStart or engageDurationMins == 0: 
+                            # Don't want to see this array entry again:
+                            dateAndSessionLenArr.pop(0)
                             continue
+                        if eventDateTime > weekEnd:
+                            # Since session start times in the array are
+                            # sorted, no need to continue going through 
+                            # the array if session date > end of the 
+                            # currently considered week. Note that we
+                            # do *not* pop the first element of the
+                            # array, so that it will be picked up next time:
+                            break
                         thisWeekThisStudentSessionList.append(engageDurationMins)
+                        # Don't want to see this array entry again:
+                        dateAndSessionLenArr.pop(0)
+                        
                     if len(thisWeekThisStudentSessionList) == 0:
                         continue
                     # Got all session lengths of this student this week:
@@ -486,32 +525,35 @@ class EngagementComputer(object):
         
         
         
-    def getCourseRuntime(self, courseName, testOnly=False):
-        
-        # Already provided an error msg for this course name?
-        if courseName in self.runtimesNotFoundCourses:
-            return(None,None)
-        
+    def getCourseRuntime(self, courseName):
+        '''
+        Query Edx.EventXtract for the earliest and latest events in the
+        given course.
+        @param courseName: name of course whose times are to be found
+        @type courseName: String
+        @return: Two-tuple with start and end time. May be (None, None) if times 
+            could not be found
+        @rtype: (datetime, datetime)
+        '''
         try:
             try:
-                runtimeLookupDb = MySQLDB(host=self.dbHost, user=self.mySQLUser, passwd=self.mySQLPwd, db=self.dbName)
+                runtimeLookupDb = MySQLDB(host=self.dbHost, user=self.mySQLUser, passwd=self.mySQLPwd, db=EngagementComputer.EVENT_XTRACT_TABLE_DB)
             except Exception as e:
                 self.logErr('While looking up course start/end times in getCourseRuntime(): %s' % `e`)
-                return (None,None)
-            if testOnly:
-                # Just ensure that the 'CourseRuntimes' table exists so
-                # that we can fail early:
-                try:
-                    runtimeLookupDb.query("SELECT course_start_date, course_end_date FROM CourseRuntimes LIMIT 1;")
-                    return
-                except Exception as e:
-                    raise ValueError('Cannot read CourseRuntimes table: %s' % `e`)
+                return(None,None)
                 
-            for runtimes in runtimeLookupDb.query("SELECT course_start_date, course_end_date FROM CourseRuntimes WHERE course_display_name = '%s';" % courseName):
+            # Get start/end time via earliest/latest observed events for given course:
+            query = 'SELECT MIN(time) AS course_start_date,\
+                            MAX(time) AS course_end_date\
+                            FROM EventXtract\
+                            WHERE course_display_name = "%s";' % courseName
+            
+            for runtimes in runtimeLookupDb.query(query):
+                if runtimes is None:
+                    return(None,None)
                 return (runtimes[0], runtimes[1])
             # No start/end times found for this course:
             self.logErr("Did not find start/end times for class '%s'" % courseName)
-            self.runtimesNotFoundCourses.append(courseName)
             return (None,None)
         except Exception as e:
             self.logErr("While attempting lookup of course start/end times: '%s'" % `e`)
@@ -577,12 +619,13 @@ class EngagementComputer(object):
                 outFileSummary.write('platform,course_display_name,TotalStudentSessions,TotalEffortAllStudents,MedPerWeekOneToTwenty,MedPerWeekTwentyoneToSixty,MedPerWeekGreaterSixty\n')
                 for className in self.classStats.keys():
                     output = 'OpenEdX,' + className + ',' + re.sub(r'[\s()]','',str(self.classStats[className]))
-                    outFileSummary.write(output + '\n') 
+                    outFileSummary.write(output + '\n')
+                outFileSummary.flush()
                 # Big detail file    
                 outFileAll.write('Platform,Course,Student,Date,Time,SessionLength\n')
                 for csvSessionRecord in self.allDataIterator():
                     outFileAll.write('OpenEdX,' + csvSessionRecord)
-                    
+                outFileAll.flush()    
                 # Student weekly effort summary:
                 outFileWeeklyEffort.write('platform,course,student,week,effortMinutes\n')
                 # For all dicts of form {student1->[[weekNum0,xMins],[weekNum1,yMins],...,],
@@ -594,7 +637,11 @@ class EngagementComputer(object):
                         # For this student get array of weekNum/time pairs:
                         studentWeeklyEffort = studentWeeklyEffortDict[student]
                         for weekNumEffortPair in studentWeeklyEffort:
-                            outFileWeeklyEffort.write('OpenEdX,%s,%s,%d,%d\n' % (course,student,weekNumEffortPair[0],weekNumEffortPair[1]))
+                            # Weeks up to this point have been zero-based.
+                            # Add one to the course week-number to make 
+                            # it 1-based:
+                            outFileWeeklyEffort.write('OpenEdX,%s,%s,%d,%d\n' % (course,student,weekNumEffortPair[0]+1,weekNumEffortPair[1]))
+                outFileWeeklyEffort.flush()
         finally:
             outFileSummary.close()
             outFileAll.close()
@@ -612,25 +659,94 @@ class EngagementComputer(object):
 if __name__ == '__main__':
     
     # -------------- Manage Input Parameters ---------------
-    if len(sys.argv) < 2:
-        sys.stderr.write('Usage: engagement.py <year as YYYY [courseName]\n')
-        sys.exit()
-    yearToProfile = sys.argv[1]
-    if len(sys.argv) > 2:
-        courseToProfile = sys.argv[2]
+    
+    usage = 'Usage: engagement.py [{courseName | None} [<year-as-YYYY>, <year-as-YYYY>, ...]]\n'
+
+    parser = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]), formatter_class=argparse.RawTextHelpFormatter)
+    parser.add_argument('-u', '--user',
+                        action='store',
+                        help='User ID that is to log into MySQL. Default: the user who is invoking this script.')
+    parser.add_argument('-p', '--pwd',
+                        action='store_true',
+                        help='Request to be asked for pwd for operating MySQL;\n' +\
+                             '    default: content of scriptInvokingUser$Home/.ssh/mysql if --user is unspecified,\n' +\
+                             '    or, if specified user is root, then the content of scriptInvokingUser$Home/.ssh/mysql_root.'
+                        )
+    parser.add_argument('-w', '--password',
+                        action='store',
+                        help='User explicitly provided password to log into MySQL.\n' +\
+                             '    default: content of scriptInvokingUser$Home/.ssh/mysql if --user is unspecified,\n' +\
+                             '    or, if specified user is root, then the content of scriptInvokingUser$Home/.ssh/mysql_root.'
+                        )
+    parser.add_argument('course',
+                        action='store',
+                        help='The course for which engagement is to be computed. Else: engagement for all courses.\n' +\
+                             "To have engagement computed for all courses, use All"
+                        ) 
+    
+    parser.add_argument('years',
+                        nargs='+',
+                        type=int,
+                        help='A list of start years (YYYY) to limit the courses that are computed. Use All if all start years are acceptable'
+                        ) 
+    
+    
+    args = parser.parse_args();
+    if args.user is None:
+        user = getpass.getuser()
     else:
-        courseToProfile = None
+        user = args.user
+        
+    if args.password and args.pwd:
+        raise ValueError('Use either -p, or -w, but not both.')
+        
+    if args.pwd:
+        pwd = getpass.getpass("Enter %s's MySQL password on localhost: " % user)
+    elif args.password:
+        pwd = args.password
+    else:
+        # Try to find pwd in specified user's $HOME/.ssh/mysql
+        currUserHomeDir = os.getenv('HOME')
+        if currUserHomeDir is None:
+            pwd = None
+        else:
+            # Don't really want the *current* user's homedir,
+            # but the one specified in the -u cli arg:
+            userHomeDir = os.path.join(os.path.dirname(currUserHomeDir), user)
+            try:
+                if user == 'root':
+                    with open(os.path.join(currUserHomeDir, '.ssh/mysql_root')) as fd:
+                        pwd = fd.readline().strip()
+                else:
+                    with open(os.path.join(userHomeDir, '.ssh/mysql')) as fd:
+                        pwd = fd.readline().strip()
+            except IOError:
+                # No .ssh subdir of user's home, or no mysql inside .ssh:
+                pwd = ''
+    
+    if args.course == 'All':
+        courseName = None
+    else:
+        courseName = args.course
+    
+    if args.years == 'All':
+        years = None
+    else:
+        years = args.years
+        
+    #**********
+#     print('courseName: %s' % str(courseName))
+#     print('years: %s' % str(years))
+#     sys.exit()
+    #**********
         
     # -------------- Run the Computation ---------------
-    #***** Switch between testing and real:
-    testing = True
-    #testing = False
-    if testing:
-        db = 'test'
-    else:
-        db = 'Misc'
+
+    db = 'Misc'
     invokingUser = getpass.getuser()
-    comp = EngagementComputer([int(yearToProfile)], 'localhost', db, 'Activities', mySQLUser=invokingUser, mySQLPwd=None, courseToProfile=courseToProfile)
+    # Set mysql password to None, which will cause
+    # the __init__() method to check ~/.ssh... 
+    comp = EngagementComputer(years, 'localhost', db, 'Activities', mySQLUser=invokingUser, mySQLPwd=None, courseToProfile=courseName)
     comp.run()
     
     # -------------- Output Results to Disk ---------------
